from __future__ import annotations
import tomli
import os
import pathlib
import wisepy2
import json

EMPTY_CONFIG = \
r"""
[julia]
min-latency = false
quiet-start = false
no-startup-file = false
interactive = true

# activate the specified project
project = ""

# this path specifies the path to your sysimage
sysimage = ""

# start Julia with loading these modules
using = []

# start Julia with executing these files
files = []
"""

def get_bool(conf: dict, name: str) -> bool | None:
    if conf.get(name, name):
        var = conf[name]
        if not isinstance(var, bool):
            raise ValueError("'{}' must be a boolean".format(name))
        return var

def get_int(conf: dict, name: str) -> int | None:
    if conf.get(name, name):
        var = conf[name]
        if not isinstance(var, int):
            raise ValueError("'{}' must be a int".format(name))
        return var

def get_str_list(conf: dict, name: str) -> list[str] | None:
    if conf.get(name, name):
        var = conf[name]
        if not isinstance(var, list) and not (all(isinstance(e, str) for e in var)):
            raise ValueError("'{}' must be a string list".format(name))
        return var

def get_str(conf: dict, name: str) -> str | None:
    if conf.get(name, name):
        var = conf[name]
        if not isinstance(var, str):
            raise ValueError("'{}' must be a string".format(name))
        return var

def write_empty_config_():
    cwd = pathlib.Path(os.getcwd())
    dev = cwd / "Development.toml"
    if dev.is_file():
        print(wisepy2.Yellow("Warning: Development.toml already exists"))
        return
    else:
        with dev.open("w", encoding="utf-8") as f:
            f.write(EMPTY_CONFIG)

def get_options() -> list[str]:
    cwd = pathlib.Path(os.getcwd())
    dev = cwd / "Development.toml"
    if dev.is_file():
        io = dev.open('rb')
        conf = tomli.load(io) # type: ignore
        if not isinstance(conf, dict):
            return []
        conf = conf.get("julia")
        if not isinstance(conf, dict):
            return []
    else:
        return []
    opts: list[str] = []

    if get_bool(conf, 'min-latency'):
        opts.append("--compile=min")
        opts.append("-O0")

    if project := get_str(conf, "project"):
        opts.append("--project={}".format(project))

    if get_bool(conf, "interactive"):
        opts.append("-i")

    if get_bool(conf, 'quiet-start'):
        opts.append("--quiet")

    if sysimage := get_str(conf, "sysimage"):
        opts.append("--sysimage")
        opts.append(sysimage)

    if get_bool(conf, "no-startup-file"):
        opts.append("--startup-file=no")

    if preload_modules := get_str_list(conf, 'using'):
        for each in preload_modules:
            opts.append("-e")
            opts.append("using {}".format(each))

    if preincluded_files := get_str_list(conf, 'files'):
        for file in preincluded_files:
            opts.append("-e")
            opts.append(
                'include(' +
                'raw' +json.dumps(str(cwd / file), ensure_ascii=False) +
                ')'
            )

    return opts



from __future__ import annotations
import venv
import wisepy2
import shutil
import json
import sys
from pathlib import Path

root = Path("~/.jlenvs").expanduser()


class Ops:
    @staticmethod
    def get_config(new_conf: "dict | None" = None) -> dict:
        root.mkdir(mode=0o755, parents=True, exist_ok=True)
        config_file = root.joinpath("config.json")
        if config_file.exists():
            if new_conf is None:
                return json.load(config_file.open("r", encoding="utf-8"))
            json.dump(new_conf, config_file.open("w", encoding="utf-8"), indent=4)
            return new_conf

        config = new_conf or {"current": None}
        with config_file.open("w", encoding="utf-8") as f:
            json.dump({"current": None}, f)
        return config

    @staticmethod
    def create_(
        name: str,
        upstream: str = "",
        version: str = "",
        confirm: bool = False,
        unstable: bool = False,
    ):
        from jill.install import install_julia

        envdir = root.joinpath(name)
        Ops.get_config()
        if envdir.exists():
            raise IOError(f"{envdir} already exists")
        venv.create(envdir.as_posix(), with_pip=True)
        root.mkdir(mode=0o755, parents=True, exist_ok=True)
        install_julia(
            version=version or name,
            install_dir=envdir.joinpath("julia").as_posix(),
            confirm=confirm,
            upstream=upstream,
            unstable=unstable,
            skip_symlinks=True,
            reinstall=True
        )
        print(wisepy2.Green(f"Environment {name} created at {envdir.as_posix()}."))

    @staticmethod
    def env(name: str):
        Ops.get_config()
        envdir = root.joinpath(name)
        if not envdir.exists():
            raise IOError(f"{envdir} does not exist")
        return envdir.as_posix()

    @staticmethod
    def list():
        Ops.get_config()
        for each in root.iterdir():
            if each.is_dir():
                yield each

    @staticmethod
    def remove_(name: str):
        config = Ops.get_config()
        envdir = root.joinpath(name)
        if not envdir.exists():
            raise IOError(f"{envdir} does not exist")
        shutil.rmtree(envdir.as_posix())
        if config.get("current") == name:
            config["current"] = None
            Ops.get_config(config)
        print(wisepy2.Green(f"Environment {name} removed."))



from __future__ import annotations
from jillw import Ops
from wisepy2 import wise
from pathlib import Path
import os
import contextlib
import io
import sys
import subprocess
import wisepy2
import shlex


def find_julia_binary_dir(env: Path):
    found = next(env.joinpath("julia").iterdir(), None)
    if found is None:
        raise IOError(f"No Julia installation found at {env}")
    return found.joinpath("bin")


@contextlib.contextmanager
def cmd_session():
    try:
        yield
    except IOError as e:
        print(wisepy2.Red("{}: {}".format(type(e).__name__, str(e))))
        sys.exit(1)


@contextlib.contextmanager
def expect_no_stdout():
    sio = io.StringIO()
    try:
        with contextlib.redirect_stdout(sio):
            yield
    finally:
        msg = sio.getvalue()
        print(wisepy2.Red(msg))

def append_PATH(PATH: str, *paths: Path):
    return os.pathsep.join([
        *map(os.path.normpath, paths),
        *PATH.split(os.pathsep)
    ])

def run_with_activated_env(cmd: list[str]):
    config = Ops.get_config()
    current: str | None = config["current"]

    if current is None:
        print(wisepy2.Yellow("No activated environment"))
        return

    env = Path(Ops.env(current))
    jlbindir = find_julia_binary_dir(env)

    if cmd and cmd[0] == "julia":
        if os.name == "nt":
            cmd[0] = jlbindir.joinpath("julia.exe").as_posix()
        else:
            cmd[0] = jlbindir.joinpath("julia").as_posix()

    if os.name == "nt":
        envdict = os.environ.copy()
        envdict["VIRTUAL_ENV"] = str(env)
        try:
            del envdict["PYTHONHOME"]
        except KeyError:
            pass
        envdict['PATH'] = append_PATH(os.environ["PATH"], jlbindir, env, env / "Scripts")
    else:
        envdict = os.environ.copy()
        envdict["VIRTUAL_ENV"] = str(env)
        try:
            del envdict["PYTHONHOME"]
        except KeyError:
            pass
        envdict['PATH'] = append_PATH(os.environ["PATH"], jlbindir, env, env / "bin")

    subprocess.run(cmd, env=envdict, shell=False)

class Main:
    @staticmethod
    def switch(name: str):
        """Switch to the specified environment"""
        with cmd_session():
            env = Ops.env(name)
            config = Ops.get_config()
            config["current"] = name
            Ops.get_config(config)
            print(wisepy2.Green(f"Switched to {name} at {env}"))

    @staticmethod
    def run(cmd: str):
        """Run a command in the activated environment"""
        run_with_activated_env(shlex.split(cmd))

    @staticmethod
    def create(
        name: str,
        version: str = "",
        upstream: str = "",
        confirm: bool = False,
        unstable: bool = False,
    ):
        """Create a new Julia environment
        """
        with cmd_session():
            Ops.create_(name, upstream, version, confirm, unstable)

    @staticmethod
    def remove(name: str):
        """Remove a Julia environment
        """
        with cmd_session():
            Ops.remove_(name)

    @staticmethod
    def list():
        """List all Julia environments
        """
        with cmd_session():
            for each in Ops.list():
                print(wisepy2.Purple(each.name, "=>", each.as_posix(), sep=" "))

    @staticmethod
    def devhere():
        """Create a Development.toml at the current directory.
        """
        with cmd_session():
            from . configloader import write_empty_config_
            write_empty_config_()

def main():
    wise(Main)()

def julia():
    extra_options: list[str] = []

    if os.environ.get("DEV", "").strip():
        from jillw.configloader import get_options
        extra_options = get_options()

    run_with_activated_env(["julia", *extra_options, *sys.argv[1:]])
